/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
 * @author Mattia Gheda <ghedamat@dei.unipd.it>
 * @author Giulio Ministeri <minister@dei.unipd.it>
 */

#ifndef MEMORY_1F_C
#define MEMORY_1F_C

#ifdef MEMORY_SIZE 
#define MAX_SIZE MEMORY_SIZE
#else
#define MAX_SIZE SLP_MEMORY_SIZE
#endif
#define MEM_HEADER_SIZE 4

typedef struct {
	uint32_t mem_id:8;
	uint32_t phy_size:13; 
	uint32_t vir_size: 11;
} __attribute__((packed))mem_header_t;

typedef uint8_t memory_id_t;
typedef uint16_t memory_size_t;

uint8_t mem_buf[MAX_SIZE];
mem_header_t* mem_buf_tail;

#ifdef SLP_REMEMBERED_BUFFER
mem_header_t* remembered_free;
#endif

#ifdef SLP_CACHE	
memory_id_t cached_id;
void* cached_p;
memory_size_t cached_size;
#endif

memory_id_t cur_mem_id = 0; 

error_t mem_free(memory_id_t id);

enum smartAlloc {
	LAYER_UNDEF = 0,
	LAYER_COAP = 70,
	LAYER_UDP = 56,
	LAYER_ICMP = 52,
	LAYER_IP = 48,
	IPV6_RX = 8
};




void mem_init ()	{
	mem_buf_tail =(mem_header_t*)( mem_buf + MAX_SIZE - MEM_HEADER_SIZE);
	mem_buf_tail->mem_id = 0;
	mem_buf_tail->phy_size = MAX_SIZE - MEM_HEADER_SIZE;
	mem_buf_tail->vir_size = 0;
	cur_mem_id = 1;
	
#ifdef SLP_REMEMBERED_BUFFER
	remembered_free = mem_buf_tail;
#endif

}


mem_header_t* find_by_id (memory_id_t id)	{
	void* tmp = mem_buf_tail;
	mem_header_t* header = mem_buf_tail;
	
	while (header->mem_id != id )	{
		tmp -= (header->phy_size + MEM_HEADER_SIZE);
		header = (mem_header_t*) tmp;

		if ((uint8_t*)tmp <= mem_buf)
			return 0;
	}
	return header;
}


memory_id_t mem_alloc (int16_t size, uint8_t layer)	{
	void* tmp;
	mem_header_t* header;
	mem_header_t* tmph;
	uint16_t hsize;

#ifdef SLP_REMEMBERED_BUFFER	
	header = remembered_free;
#else
	header = mem_buf_tail;
#endif

	tmp = mem_buf_tail;
	
	while (header->mem_id != 0 || header->phy_size < (size+layer))	{
		tmp -= header->phy_size + MEM_HEADER_SIZE;
		header = (mem_header_t*) tmp;
		if ((uint8_t*)tmp <= mem_buf)
			return 0;
	}
	
	hsize = size+layer+MEM_HEADER_SIZE;
						
	if (header->phy_size >= hsize)	{
		tmp -= hsize;
		if ((uint8_t*)tmp > mem_buf)	{
			tmph = (mem_header_t*) tmp;
			tmph->mem_id = 0;
			tmph->phy_size = header->phy_size - hsize;
			header->phy_size = size + layer;
		}
	}			
	
	while (find_by_id(cur_mem_id) != NULL)
		cur_mem_id++;
		
	header->mem_id = cur_mem_id++;
	if (cur_mem_id == 0)
		cur_mem_id++;
			
	header->vir_size = size;

#ifdef SLP_CACHE	
	cached_id = header->mem_id;
	cached_p = header;
	cached_p -= header->vir_size;
	cached_size = header->vir_size;
#endif
	
	return header->mem_id;
}

void* mem_id2p (memory_id_t id, memory_size_t* size)	{
	void* tmp;
	mem_header_t* header;
	
#ifdef SLP_CACHE	
	if (cached_id == id)	{
		*size = cached_size;
		return cached_p;
	}
#endif

	*size = 0;
	header = find_by_id(id);
	if (!header)
		return NULL;
		
	tmp = header;
	tmp -= header->vir_size;
	*size = header->vir_size;

#ifdef SLP_CACHE		
	cached_id = id;
	cached_p = tmp;
	cached_size = header->vir_size;
#endif
	return tmp;
}


unsigned char mem_hrealloc (memory_id_t id, int16_t size)	{
	void* new_mem_buf;
	void* old_mem_buf;
	memory_size_t new_sz, old_sz;
	memory_id_t new_id;
	mem_header_t* header = find_by_id(id);
	
	if (!header)
		return 1;
		
	if (0 >= size)	{
		if ( -size > header->vir_size)
			return -1;
		
		else	{
#ifdef SLP_CACHE	
			if (cached_id == id)	{
				cached_p -= size;
				cached_size += size;
			}		
#endif

			header->vir_size += size;
		}
	}
	
	else if ((header->phy_size - header->vir_size) >= size )	{
#ifdef SLP_CACHE		
		if (cached_id == id)	{
			cached_p -= size;
			cached_size += size;
		}		
#endif

		header->vir_size += size;
	}	
		
	else	{
		new_id = mem_alloc (header->vir_size + size, LAYER_UNDEF);
		if (new_id == 0)	{			
			return 1;
			}
			
		new_mem_buf = mem_id2p(new_id, &new_sz);
		old_mem_buf = mem_id2p(id, &old_sz);
		
		header = (mem_header_t*) (new_mem_buf + new_sz);
		memcpy (new_mem_buf + size, old_mem_buf, old_sz);
		mem_free(id);
		header->mem_id = id;
#ifdef SLP_CACHE	
		if (cached_id == id)	{
			cached_p = new_mem_buf;
			cached_size = new_sz;
		}	
#endif
	}
	return 0;
}

error_t mem_free (memory_id_t id)	{
	void* tmp = mem_buf_tail;
	mem_header_t* prev_header = NULL;
	mem_header_t* next_header = NULL;
	mem_header_t* header = mem_buf_tail;

#ifdef SLP_CACHE	
	if (cached_id == id)
		cached_id = 0;
#endif
		
	
	while (header->mem_id != id )	{
		tmp -= (header->phy_size + MEM_HEADER_SIZE);
		prev_header = header;
		header = (mem_header_t*) tmp;

		if ((uint8_t*)tmp <= mem_buf)
			return 1;
	}	
	
	header->mem_id = 0;
	
	tmp -= (header->phy_size + MEM_HEADER_SIZE);
	
	if ((uint8_t*)tmp <= mem_buf)
		next_header = NULL;
	else
		next_header = (mem_header_t*) tmp;
	
	if (next_header == NULL && prev_header == NULL)
		mem_init();

	else if (prev_header != NULL && prev_header->mem_id == 0)	{

#ifdef SLP_REMEMBERED_BUFFER		
			if (remembered_free < prev_header)
				remembered_free = prev_header;
#endif

			prev_header->phy_size += header->phy_size + MEM_HEADER_SIZE;
			if (next_header != NULL && next_header->mem_id == 0)
				prev_header->phy_size += next_header->phy_size + MEM_HEADER_SIZE;
	}

	else if (next_header != NULL && next_header->mem_id == 0)	{

#ifdef SLP_REMEMBERED_BUFFER
			if (remembered_free < header)
				remembered_free = header;
#endif
				
			header->phy_size += next_header->phy_size + MEM_HEADER_SIZE;
	}
	
	return 0;
}


#ifdef PC_TEST
void mem_dump (void)	{
	#define EMPTY 31 
	#define HEADER 32
	#define VIR_SIZE 33
	#define PHY_SIZE 34
	#define COMMENT 35
	
	int i;
	unsigned char* tmp = (unsigned char*) mem_buf_tail;
	mem_header_t* header = mem_buf_tail;
	
	while (tmp >= mem_buf)	{
		printf("\n");
		if (header->mem_id == 0)
			printf("%c[%d;%d;%dmempty block size %d ", 0x1B, 0,COMMENT, 10, header->phy_size);
		else
			printf("%c[%d;%d;%dmblock id: %d phy size %d, vir size %d ", 0x1B, 0,COMMENT, 10, header->mem_id, header->phy_size, header->vir_size);
		 
		for (i=0; i<MEM_HEADER_SIZE; i++)
			printf("%c[%d;%d;%dm%x ", 0x1B, 0,HEADER, 10, *(tmp+i));
		
		if (header->mem_id != 0)	{
			tmp -= header->vir_size;
			for (i=0; i<header->vir_size; i++)	
				printf("%c[%d;%d;%dm%x ", 0x1B, 0,VIR_SIZE, 10, *(tmp+i));
			
			tmp -= header->phy_size - header->vir_size;
			for (i=0; i<(header->phy_size-header->vir_size); i++)	
				printf("%c[%d;%d;%dm%x ", 0x1B, 0,PHY_SIZE, 10, *(tmp+i));
		}
		else	{
			tmp -= header->phy_size;
			for (i=0; i<header->phy_size; i++)	
					printf("%c[%d;%d;%dm%x ", 0x1B, 0,EMPTY, 10, *(tmp+i));	
		}
		
		tmp -= MEM_HEADER_SIZE;	
		header = (mem_header_t*) tmp;
	}
	
	printf("%c[%dm", 0x1B, 0);
	printf("\n");
}	
		
#endif		

#endif
